((c++-mode . (
  (c-file-style . "k&r")
  (c-basic-offset . 4)
  (indent-tabs-mode . nil)
  (tab-width . 4)
  (show-trailing-whitespace . t)
  (indicate-empty-lines . t)
  (eval . (c-set-offset 'innamespace 0))
  (eval . (c-set-offset 'defun-open 0))
  (eval . (c-set-offset 'inline-open 0))
  (eval . (c-set-offset 'arglist-intro '+))
  (eval . (c-set-offset 'arglist-cont 0))
  (eval . (c-set-offset 'arglist-cont-nonempty '+))
  (eval . (c-set-offset 'substatement-open 0))
  (eval . (c-set-offset 'access-label '-))
  )))
